/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  ======== Server.c ========
 *
 */

/* this define must precede inclusion of any xdc header file */
#define Registry_CURDESC Test__Desc
#define MODULE_NAME "Server"

//#define DEBUG

/* xdctools header files */
#include <xdc/std.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Diags.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/Registry.h>

#include <stdio.h>

/* package header files */
#include <ti/ipc/MessageQ.h>
#include <ti/ipc/MultiProc.h>
#include <ti/ipc/SharedRegion.h>
#include <ti/ipc/HeapMemMP.h>
#include <ti/ipc/remoteproc/Resource.h>
#include <ti/sysbios/hal/Cache.h>
#include <xdc/runtime/IHeap.h>
#include <xdc/runtime/Memory.h>

#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>

/* local header files */
#include "../shared/AppCommon.h"

/* module header file */
#include "Server.h"
#include "audio.h"
#include "../host/ARM_audio.h"
#include <_bigdataxlat.h>
#include <math.h>

Shared_mem *shmem;

/* module structure */
typedef struct
{
    UInt16 hostProcId;        // host processor id
    MessageQ_Handle slaveQue; // created locally
} Server_Module;

/* private data */
Registry_Desc Registry_CURDESC;
static Server_Module Module;

/*
 *  ======== Server_init ========
 */
Void Server_init(Void)
{
    Registry_Result result;

    /* register with xdc.runtime to get a diags mask */
    result = Registry_addModule(&Registry_CURDESC, MODULE_NAME);
    Assert_isTrue(result == Registry_SUCCESS, (Assert_Id)NULL);

    /* initialize module object state */
    Module.hostProcId = MultiProc_getId("HOST");
}

/*
 *  ======== Server_create ========
 */
Int Server_create()
{
    Int status = 0;
    MessageQ_Params msgqParams;
    char msgqName[32];

    /* enable some log events */
    Diags_setMask(MODULE_NAME "+EXF");

    /* create local message queue (inbound messages) */
    MessageQ_Params_init(&msgqParams);
    sprintf(msgqName, App_SlaveMsgQueName, MultiProc_getName(MultiProc_self()));
    Module.slaveQue = MessageQ_create(msgqName, &msgqParams);

    if (Module.slaveQue == NULL)
    {
        status = -1;
        goto leave;
    }

    Log_print0(Diags_INFO, "Server_create: server is ready");

leave:
    Log_print1(Diags_EXIT, "<-- Server_create: %d", (IArg)status);
    return (status);
}

/*
 *  ======== Server_exec ========
 */
Int Server_exec()
{
    Int status;
    Bool running = TRUE;
    App_Msg *msg;
    MessageQ_QueueId queId;
    UInt16 regionId1 = 1;
    Uint32 *bigDataLocalPtr;
    Int j;
    UInt32 errorCount = 0;
    Int retVal;
    bigDataLocalDesc_t bigDataLocalDesc;
    SharedRegion_Entry srEntry;
    void *sharedRegionAllocPtr = NULL;

    int process_pos = 0;

    UInt32 i = 0;
    int count = 0;

    // for(i = 0; i<FRAMES_PER_BUFFER*N_CHANNELS;i++)
    // {
    //     localBuf[i] = 0;
    // }

    setupDSP(FRAMES_PER_BUFFER, SAMPLE_RATE);

    Log_print0(Diags_ENTRY | Diags_INFO, "--> Server_exec:");

    while (running)
    {
    
        //if(count>2)
        
        // MessageQ_get(Module.slaveQue, (MessageQ_Msg *)&msg,
                       //       MessageQ_FOREVER);
        /* wait for inbound message */
        status = MessageQ_get(Module.slaveQue, (MessageQ_Msg *)&msg,
                              MessageQ_FOREVER);

        
        count = MessageQ_count(Module.slaveQue);
        //Log_print1(Diags_INFO, "n messages in Q...%d", count);

        if (status < 0)
        {
            goto leave;
        }
        Log_print1(Diags_ENTRY | Diags_INFO, "Message received...count %d", count);//msg->id);
        switch (msg->cmd)
        {
        case App_CMD_SHARED_REGION_INIT:
            /* Create Shared region with information from init message */
            /* Configure srEntry */

            status = Resource_physToVirt((UInt32)msg->u.sharedRegionInitCfg.base,
                                         (UInt32 *)&sharedRegionAllocPtr);
            if (status != Resource_S_SUCCESS)
            {
                printf("Resource_physToVirt failed \n");
                goto leave;
            }
            srEntry.base = sharedRegionAllocPtr;
            srEntry.len = msg->u.sharedRegionInitCfg.size;
            srEntry.ownerProcId = MultiProc_self();
            srEntry.isValid = TRUE;
            /* Make sure CacheEnable is TRUE if using cached memory */
            srEntry.cacheEnable = TRUE;
            srEntry.createHeap = FALSE;
            srEntry.cacheLineSize = 128;
            srEntry.name = "SR1";

            status = SharedRegion_setEntry(regionId1, &srEntry);
            Log_print0(Diags_ENTRY | Diags_INFO, "Shared region entry configured...");

            break;
        case App_CMD_NEWDATA:
            break;

        case App_CMD_BIGDATA:
#ifdef DEBUG
            Log_print1(Diags_ENTRY | Diags_INFO, "msg->cmd=App_CMD_BIGDATA,msg->ptr=0x%x",
                       (IArg)msg->u.bigDataSharedDesc.sharedPtr);
#endif

            /* Translate to local descriptor */
            retVal = bigDataXlatetoLocalAndSync(regionId1,
                                                &msg->u.bigDataSharedDesc, &bigDataLocalDesc);
            if (retVal)
            {
                status = -1;
                goto leave;
            }
            shmem = (Shared_mem *)bigDataLocalDesc.localPtr;
#ifdef DEBUG
            /* print message from buffer */
            Log_print1(Diags_INFO, " Received message %d", msg->id);
            Log_print1(Diags_INFO, " Local Pointer 0x%x", (UInt32)bigDataLocalPtr);
            Log_print0(Diags_INFO, " First 8 bytes: ");
            for (j = 0; j < 8 && j < bigDataLocalDesc.size / sizeof(uint32_t); j += 4)
                Log_print4(Diags_INFO, "0x%x, 0x%x, 0x%x, 0x%x",
                           bigDataLocalPtr[j], bigDataLocalPtr[j + 1], bigDataLocalPtr[j + 2], bigDataLocalPtr[j + 3]);
            Log_print0(Diags_INFO, " Last 8 bytes: ");
            for (j = (bigDataLocalDesc.size / sizeof(uint32_t)) - 8;
                 j < bigDataLocalDesc.size / sizeof(uint32_t); j += 4)
                Log_print4(Diags_INFO, "0x%x, 0x%x, 0x%x, 0x%x",
                           bigDataLocalPtr[j], bigDataLocalPtr[j + 1], bigDataLocalPtr[j + 2], bigDataLocalPtr[j + 3]);
#endif
            /* Check values to see expected results */
            // for( j=0; j < shmem->Buf_size; j++) {
            //     if ( bigDataLocalPtr[j] != (msg->id+j) ) {
            //         errorCount++;
            //     }
            // }

            /* Fill new data */
            for (j = 0; j < FRAMES_PER_BUFFER; j++)
            {
                shmem->dataBuf[0][2 * j] *= 1;
                shmem->dataBuf[0][2 * j + 1] *= 0;
            }

            // audioDSPCallback(&shmem->dataBuf[0], NULL, FRAMES_PER_BUFFER);
            //  shmem->rd = (shmem->rd + 1) & (N_BUFFS - 1);

            /* Translate to Shared Descriptor and Sync */
            retVal = bigDataXlatetoGlobalAndSync(regionId1,
                                                 &bigDataLocalDesc, &msg->u.bigDataSharedDesc);
            if (retVal)
            {
                status = -1;
                goto leave;
            }
            break;

        case App_CMD_SHUTDOWN:
            running = FALSE;
            break;

        default:
            break;
        }

        /* process the message */
        Log_print2(Diags_INFO, "Server_exec: processed id %d, cmd=0x%x", msg->id, msg->cmd);

        /* send message back */
        queId = MessageQ_getReplyQueue(msg);
        MessageQ_put(queId, (MessageQ_Msg)msg);
    } /* while (running) */

leave:
    /* Print error count if non-zero */
    if (errorCount)
    {
        Log_print1(Diags_INFO, "Server_exec: Error Count %d", errorCount);
        status = -1;
    }
    else
        Log_print0(Diags_INFO, "Server_exec: Data check clean");

    Log_print1(Diags_EXIT, "<-- Server_exec: %d", (IArg)status);
    return (status);
}

/*
 *  ======== Server_delete ========
 */

Int Server_delete()
{
    Int status;

    Log_print0(Diags_ENTRY, "--> Server_delete:");

    /* delete the video message queue */
    status = MessageQ_delete(&Module.slaveQue);

    if (status < 0)
    {
        goto leave;
    }

leave:
    if (status < 0)
    {
        Log_error1("Server_finish: error=0x%x", (IArg)status);
    }

    /* disable log events */
    Log_print1(Diags_EXIT, "<-- Server_delete: %d", (IArg)status);
    Diags_setMask(MODULE_NAME "-EXF");

    return (status);
}

/*
 *  ======== Server_exit ========
 */

Void Server_exit(Void)
{
    /*
     * Note that there isn't a Registry_removeModule() yet:
     *     https://bugs.eclipse.org/bugs/show_bug.cgi?id=315448
     *
     * ... but this is where we'd call it.
     */
}
